@@ -1,24 +1,30 @@
 /*
- * Copyright (c) 2015, 2016, Oracle and/or its affiliates. All rights reserved.
- *
- * This code is licensed under the terms of the GPLv2
- * <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>, like most
- * MySQL Connectors. There are special exceptions to the terms and
- * conditions of the GPLv2 as it is applied to this software, see the
- * FLOSS License Exception
- * <http://www.mysql.com/about/legal/licensing/foss-exception.html>.
+ * Copyright (c) 2015, 2018, Oracle and/or its affiliates. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published
- * by the Free Software Foundation; version 2 of the License.
+ * it under the terms of the GNU General Public License, version 2.0, as
+ * published by the Free Software Foundation.
+ *
+ * This program is also distributed with certain software (including
+ * but not limited to OpenSSL) that is licensed under separate terms,
+ * as designated in a particular file or component or in included license
+ * documentation.  The authors of MySQL hereby grant you an
+ * additional permission to link the program and your derivative works
+ * with the separately licensed software that they have included with
+ * MySQL.
+ *
+ * Without limiting anything contained in the foregoing, this file,
+ * which is part of MySQL Connector/C++, is also subject to the
+ * Universal FOSS Exception, version 1.0, a copy of which can be found at
+ * http://oss.oracle.com/licenses/universal-foss-exception.
  *
  * This program is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
- * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
- * for more details.
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+ * See the GNU General Public License, version 2.0, for more details.
  *
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
  */
 
@@ -26,9 +32,10 @@
 #include <mysql/cdk/codec.h>
 #include "../parser/json_parser.h"
 
-PUSH_SYS_WARNINGS
+PUSH_SYS_WARNINGS_CDK
+#include <sstream>
 #include <algorithm>  // std::min
-POP_SYS_WARNINGS
+POP_SYS_WARNINGS_CDK
 
 // Include Protobuf headers needed for decoding float numbers
 
@@ -60,13 +67,8 @@ size_t cdk::Codec<TYPE_BYTES>::to_bytes(const std::string &str, bytes raw)
   return len;
 }
 
-size_t Codec<TYPE_STRING>::measure(const string &str)
-{
-  return get_codec().measure(str);
-}
 
-
-size_t Codec<TYPE_STRING>::from_bytes(bytes raw, string &str)
+size_t Codec<TYPE_STRING>::from_bytes(bytes raw, cdk::string &str)
 {
   //TODO: padding
 
@@ -84,7 +86,7 @@ size_t Codec<TYPE_STRING>::from_bytes(bytes raw, string &str)
 }
 
 
-size_t Codec<TYPE_STRING>::to_bytes(const string& str, bytes raw)
+size_t Codec<TYPE_STRING>::to_bytes(const cdk::string& str, bytes raw)
 {
   return get_codec().to_bytes(str, raw);
 }
@@ -92,18 +94,40 @@ size_t Codec<TYPE_STRING>::to_bytes(const string& str, bytes raw)
 
 foundation::api::String_codec* Format<TYPE_STRING>::codec() const
 {
+  using namespace foundation;
   /*
-    TODO: This implementation uses ASCII codec for all non utf8 strings,
-    which works only for simple strings. Correctly handle all MySQL
-    character encodings.
+    Note: Starting from 8.0.14, xplugin is always sending strings using
+    the connection encoding, which is utf8mb4. This way all MySQL charsets
+    are correctly handled as transcoding to utf8 is done in the server.
+
+    For pre-8.0.14 servers, or if connection encoding is changed somehow,
+    we might get strings in different encoding. We can still fully handle
+    Unicode based encodings. For anything else we fall back to ASCII decoder
+    which will work for strings that contain ASCII chars only - if not,
+    conversion error will be thrown.
+
+    Note: Standard endianess for multi-byte encodings is big-endian. This is
+    what MySQL server uses.
   */
 
-  static foundation::String_codec<foundation::codecvt_utf8>  utf8;
-  static foundation::String_codec<foundation::codecvt_ascii> ascii;
+  static String_codec<String_encoding::UTF8>    utf8;
+  static String_codec<String_encoding::UTF16BE> utf16;
+  static String_codec<String_encoding::UCS4BE>  ucs;
+  static String_codec<String_encoding::ASCII>   ascii;
 
-  return Charset::utf8 == charset() ?
-      (foundation::api::String_codec*)&utf8
-    : (foundation::api::String_codec*)&ascii;
+  switch (charset())
+  {
+  case Charset::utf8:
+  case Charset::utf8mb4:
+    return (foundation::api::String_codec*)&utf8;
+  case Charset::ucs2:
+  case Charset::utf16:
+    return (foundation::api::String_codec*)&utf16;
+  case Charset::utf32:
+    return (foundation::api::String_codec*)&ucs;
+  default:
+    return (foundation::api::String_codec*)&ascii;
+  }
 }
 
 
@@ -352,11 +376,80 @@ size_t Codec<TYPE_INTEGER>::to_bytes(uint64_t val, bytes buf)
   return internal_to_bytes(val, buf);
 }
 
+std::string Codec<TYPE_FLOAT>::internal_decimal_to_string(bytes buf)
+{
+  if (buf.size() < 2)
+    THROW("Invalid DECIMAL buffer");
+  byte scale_digits = *buf.begin();
+  byte sign_byte = *(buf.end() - 1);
+  int last_digit = -1;
+  bool is_negative;
+
+  /*
+    Last 4 bits of DECIMAL should always be 1100 (0xC) for positive
+    or 1101 (0xD) for negative value
+  */
+  if ((sign_byte & 0x0C) == 0x0C)
+  {
+    /* A digit must be retrieved from the first 4 bits of the sign byte */
+    last_digit = (int)(sign_byte >> 4);
+    is_negative = (sign_byte & 0x0D) == 0x0D;
+  }
+  else if ((sign_byte & 0xC0) == 0xC0)
+  {
+    /* No digit in the sign byte */
+    is_negative = (sign_byte & 0xD0) == 0xD0;
+  }
+  else
+    THROW("Invalid DECIMAL buffer");
+
+  int total_digits = ((int)buf.size() - 2) * 2 + (last_digit + 1 ? 1 : 0);
+  if (total_digits <= scale_digits)
+    THROW("Invalid DECIMAL buffer");
+
+  std::stringstream sstream;
+
+  if (is_negative)
+    sstream << "-";
+
+  int pos = 0;
+  for (byte *b = buf.begin() + 1; b < buf.end() - 1; ++b)
+  {
+    do
+    {
+      if (total_digits - scale_digits == pos)
+        /* Getting the locale decimal point */
+        sstream << std::use_facet< std::numpunct<char> >(sstream.getloc()).decimal_point();;
+
+      if (pos % 2)
+        sstream << (int)(*b & 0x0F);
+      else
+        sstream << (int)(*b >> 4);
+      ++pos;
+    } while (pos % 2);
+  }
+
+  if (last_digit + 1)
+    sstream << last_digit;
+
+  return sstream.str();
+}
+
 
 size_t Codec<TYPE_FLOAT>::from_bytes(bytes buf, float &val)
 {
   if (m_fmt.type() == cdk::Format<cdk::TYPE_FLOAT>::DECIMAL)
-    THROW("Codec<TYPE_FOAT>: DECIMAL format not supported yet");
+  {
+    std::string s = internal_decimal_to_string(buf);
+    const char *data = s.c_str();
+    char *str_end;
+    float f = strtof(data, &str_end);
+
+    if (*str_end != '\0' || f == std::numeric_limits<float>::infinity())
+      THROW("Codec<TYPE_FLOAT>: conversion overflow");
+    val = f;
+    return buf.size();
+  }
 
   if (m_fmt.type() == cdk::Format<cdk::TYPE_FLOAT>::DOUBLE)
     throw Error(cdkerrc::conversion_error,
@@ -368,7 +461,7 @@ size_t Codec<TYPE_FLOAT>::from_bytes(bytes buf, float &val)
 
   if (sz < buf.size())
     throw Error(cdkerrc::conversion_error,
-                "Codec<TYPE_FLOAT>: convertion overflow");
+                "Codec<TYPE_FLOAT>: conversion overflow");
 
   val = google::protobuf::internal::WireFormatLite::DecodeFloat(val_tmp);
   return sz;
@@ -378,7 +471,18 @@ size_t Codec<TYPE_FLOAT>::from_bytes(bytes buf, float &val)
 size_t Codec<TYPE_FLOAT>::from_bytes(bytes buf, double &val)
 {
   if (m_fmt.type() == cdk::Format<cdk::TYPE_FLOAT>::DECIMAL)
-    THROW("Codec<TYPE_FOAT>: DECIMAL format not supported yet");
+  {
+    std::string s = internal_decimal_to_string(buf);
+    const char *data = s.c_str();
+    char *str_end;
+    double d = strtod(data, &str_end);
+
+    /* No need to check for value overflow from DECIMAL to double */
+    if (*str_end != '\0')
+      THROW("Codec<TYPE_FLOAT>: conversion overflow");
+    val = d;
+    return buf.size();
+  }
 
   size_t sz;
 
@@ -395,7 +499,7 @@ size_t Codec<TYPE_FLOAT>::from_bytes(bytes buf, double &val)
 
   if (sz < buf.size())
     throw Error(cdkerrc::conversion_error,
-                "Codec<TYPE_FLOAT>: convertion overflow");
+                "Codec<TYPE_FLOAT>: conversion overflow");
 
   val = google::protobuf::internal::WireFormatLite::DecodeDouble(val_tmp);
   return sz;
@@ -423,7 +527,7 @@ size_t Codec<TYPE_FLOAT>::to_bytes(double val, bytes buf)
 size_t Codec<TYPE_DOCUMENT>::from_bytes(bytes data, JSON::Processor &jp)
 {
   std::string json_string(data.begin(), data.end());
-  JSON_parser parser(json_string);
+  JSON_parser parser(std::move(json_string));
   parser.process(jp);
   return 0; // FIXME
 }
