@@ -1,24 +1,30 @@
 /*
- * Copyright (c) 2015, 2016, Oracle and/or its affiliates. All rights reserved.
- *
- * This code is licensed under the terms of the GPLv2
- * <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>, like most
- * MySQL Connectors. There are special exceptions to the terms and
- * conditions of the GPLv2 as it is applied to this software, see the
- * FLOSS License Exception
- * <http://www.mysql.com/about/legal/licensing/foss-exception.html>.
+ * Copyright (c) 2015, 2020, Oracle and/or its affiliates.
  *
  * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published
- * by the Free Software Foundation; version 2 of the License.
+ * it under the terms of the GNU General Public License, version 2.0, as
+ * published by the Free Software Foundation.
+ *
+ * This program is also distributed with certain software (including
+ * but not limited to OpenSSL) that is licensed under separate terms,
+ * as designated in a particular file or component or in included license
+ * documentation.  The authors of MySQL hereby grant you an
+ * additional permission to link the program and your derivative works
+ * with the separately licensed software that they have included with
+ * MySQL.
+ *
+ * Without limiting anything contained in the foregoing, this file,
+ * which is part of MySQL Connector/C++, is also subject to the
+ * Universal FOSS Exception, version 1.0, a copy of which can be found at
+ * http://oss.oracle.com/licenses/universal-foss-exception.
  *
  * This program is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
- * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
- * for more details.
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+ * See the GNU General Public License, version 2.0, for more details.
  *
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write to the Free Software Foundation, Inc.,
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
  */
 
@@ -30,6 +36,7 @@
 // Note: on Windows this includes windows.h
 
 #include <mysql/cdk/foundation/common.h>
+#include <google/protobuf/io/zero_copy_stream.h>
 
 /*
   Note: On Windows the INIT_ONCE structure was added only in later
@@ -47,16 +54,16 @@
 
 #include "protocol.h"
 
-PUSH_SYS_WARNINGS
+PUSH_SYS_WARNINGS_CDK
 #include <memory.h> // for memcpy
-POP_SYS_WARNINGS
+POP_SYS_WARNINGS_CDK
 
 
 #ifdef DEBUG_PROTOBUF
 
-PUSH_SYS_WARNINGS
+PUSH_SYS_WARNINGS_CDK
 #include <iostream>
-POP_SYS_WARNINGS
+POP_SYS_WARNINGS_CDK
 
 #define MSG_CLIENT_name_client(MSG,N,C) case msg_type::cli_##N: return #MSG;
 #define MSG_SERVER_name_client(MSG,N,C)
@@ -113,9 +120,9 @@ namespace mysqlx {
 static void log_handler(LogLevel level, const char* filename, int line, const std::string& message);
 
 #ifdef _WIN32
-BOOL CALLBACK log_handler_init( PINIT_ONCE InitOnce, PVOID Parameter, PVOID *lpContext)
+BOOL CALLBACK log_handler_init(PINIT_ONCE, PVOID, PVOID*)
 {
-  SetLogHandler(log_handler);
+  SetLogHandler(&log_handler);
   return TRUE;
 }
 #else
@@ -137,19 +144,23 @@ Protocol_impl::Protocol_impl(Protocol::Stream *str, Protocol_side side)
   , m_msg_state(PAYLOAD)
   , m_msg_size(0)
 {
-  EXECUTE_ONCE(&log_handler_once, &log_handler_init);
+  // Warning can be disabled because the handler is not called, only registered
+  PUSH_MSVC17_WARNINGS_CDK
+    EXECUTE_ONCE(&log_handler_once, &log_handler_init);
+  POP_MSVC17_VARNINGS_CDK
 
-  // Allocate initial I/O buffers
+    // Allocate initial I/O buffers
 
-  m_wr_size= m_rd_size= 512;
-  m_rd_buf= (byte*)malloc(m_rd_size);
-  m_wr_buf= (byte*)malloc(m_wr_size);
+  m_wr_size = m_rd_size = 1024;
+  m_rd_buf = (byte*)malloc(m_rd_size);
+  m_wr_buf = (byte*)malloc(m_wr_size);
 
   if (!m_rd_buf)
     throw_error("Could not allocate initial input buffer");
 
   if (!m_wr_buf)
     throw_error("Could not allocate initial output buffer");
+
 }
 
 Protocol_impl::~Protocol_impl()
@@ -159,7 +170,6 @@ Protocol_impl::~Protocol_impl()
   delete m_str;
 }
 
-
 class Invalid_msg_error : public Error_class<Invalid_msg_error>
 {
   unsigned m_state;
@@ -181,6 +191,24 @@ private:
   }
 };
 
+void Protocol_impl::start_Pipeline()
+{
+  m_pipeline = true;
+}
+
+Protocol::Op& Protocol_impl::snd_Pipeline()
+{
+  m_snd_op.reset();
+  m_snd_op.reset(new Op_snd_pipeline(*this));
+  return *m_snd_op;
+}
+
+void Protocol_impl::clear_Pipeline()
+{
+  m_pipeline = false;
+  m_pipeline_size = 0;
+}
+
 
 Protocol::Op& Protocol_impl::snd_start(Message &msg, msg_type_t msg_type)
 {
@@ -191,7 +219,7 @@ Protocol::Op& Protocol_impl::snd_start(Message &msg, msg_type_t msg_type)
   using std::endl;
 
   cerr << endl;
-  cerr << ">>>> Seding message >>>>" << endl;
+  cerr << ">>>> Sending message >>>>" << endl;
   cerr << "of type " << msg_type << ": "
       << msg_type_name(CLIENT, msg_type) << endl;
   cerr << msg.DebugString();
@@ -261,7 +289,10 @@ Message* mk_message(Protocol_side side, msg_type_t msg_type)
   situation occurs in Protobuf (such as parsing error etc).
 */
 
-static void log_handler(LogLevel level, const char* filename, int line, const std::string& message)
+static void log_handler(
+  LogLevel level, const char* /*filename*/, int /*line*/,
+  const std::string& message
+)
 {
   switch(level)
   {
@@ -290,44 +321,152 @@ static void log_handler(LogLevel level, const char* filename, int line, const st
   }
 }
 
+/*
+  Implementation of protobuf's ZeroCopyOutputStream which stores
+  data in the given memory buffer.
+  ==============================
+*/
+
+class ArrayStream : public google::protobuf::io::ZeroCopyOutputStream
+{
+  byte *m_buf;
+  size_t m_size;
+  size_t m_bytes_count;
+
+  public:
+
+  ArrayStream(byte * buf, size_t size) : m_buf(buf), m_size(size),
+                                         m_bytes_count(0)
+  {}
+
+  virtual bool Next(void ** data, int * size)
+  {
+    if (m_bytes_count >= m_size)
+      return false;
+
+    *data = m_buf + m_bytes_count;
+    *size = (int)(m_size - m_bytes_count);
+    m_bytes_count = m_size; // We always guess that all buffer is used
+
+    return true;
+  }
+
+  virtual void BackUp(int count)
+  {
+    assert((int)m_bytes_count >= count);
+    m_bytes_count -= count;
+  }
+
+  int64 ByteCount() const
+  {
+    return (int64)m_bytes_count;
+  }
+};
 
 /*
   Writing/reading message frames
   ==============================
 */
 
-
 void Protocol_impl::write_msg(msg_type_t msg_type, Message &msg)
 {
   if (m_wr_op)
-    THROW("Can't write message while another one is written");
+    THROW("Can't write message while another one is being written");
 
   msg_size_t net_size = static_cast<unsigned>(msg.ByteSize()) + 1;
 
   if (!resize_buf(CLIENT, header_length + net_size))
     THROW("Not enough memory for output buffer");
 
-  // Construct message header
-
-  HTONSIZE(net_size);
-  memcpy((void*)m_wr_buf, (const void*)&net_size, sizeof(net_size));
-  m_wr_buf[header_length - 1] = (byte)msg_type;
-
-  // Convert net_size back to original endian before using it later
-
-  NTOHSIZE(net_size);
-
   // Serialize message
 
   assert(m_wr_size < (size_t)std::numeric_limits<int>::max());
 
-  if (!msg.SerializeToArray((void*)(m_wr_buf + header_length),
-                            (int)(m_wr_size - header_length)))
+  if (!msg.SerializeToArray((void*)(wr_buffer() + header_length),
+                            (int)(wr_size() - header_length)))
+  {
+    m_pipeline = false;
+    m_pipeline_size = 0;
     throw_error(cdkerrc::protobuf_error, "Serialization error!");
+  }
+
+  byte *wr_buf = wr_buffer();
+  size_t total_write_size = 0;
+
+  if (m_compressor.m_compression_type != Compression_type::NONE &&
+      net_size > m_compress_threshold)
+  {
+    HTONSIZE(net_size);
+    memcpy((void*)wr_buf, (const void*)&net_size, sizeof(net_size));
+    NTOHSIZE(net_size);
+    wr_buf[header_length - 1] = (byte)msg_type;
+
+    // Do not take into account the msg type when using compression
+    msg_size_t payload_size = net_size - 1;
+    msg_size_t compressed_size = (msg_size_t)m_compressor.
+      do_compress(m_wr_buf, payload_size + header_length);
+
+    if (compressed_size == 0)
+      throw_error("Failed to compress the data");
+
+    /*
+      Two messages are required in order to ensure that
+      the message type and uncompressed size are sent before
+      the payload.
+    */
+    Mysqlx::Connection::Compression first_fields;
+    Mysqlx::Connection::Compression compression_payload;
+
+    first_fields.set_client_messages(
+      static_cast<::Mysqlx::ClientMessages_Type>(msg_type));
+    first_fields.set_uncompressed_size(payload_size + header_length);
+    byte *cmp_out_buf = m_compressor.get_out_buf();
+    compression_payload.set_payload(cmp_out_buf, compressed_size);
+
+    // The Compressed message will add only a few bytes
+    // to the compressed payload. It should not be more than 128.
+    if (!resize_buf(CLIENT, compressed_size + 128))
+      THROW("Not enough memory for output buffer");
+
+    wr_buf = wr_buffer();
+
+    ArrayStream astr(wr_buf + header_length, wr_size());
+    first_fields.SerializePartialToZeroCopyStream(&astr);
+    compression_payload.SerializePartialToZeroCopyStream(&astr);
+
+    // First 4 bytes of frame length are not counted as payload
+    net_size = static_cast<msg_size_t>(astr.ByteCount()) + 1;
+    msg_type = msg_type::cli_Compression;
+
+  }
+
+  // Construct message header
+  HTONSIZE(net_size);
+  memcpy((void*)wr_buf, (const void*)&net_size, sizeof(net_size));
+  wr_buf[header_length - 1] = (byte)msg_type;
+  // Convert net_size back to original endian before using it later
+  NTOHSIZE(net_size);
+  total_write_size = net_size + header_length - 1;
 
   // Create write operation to send message payload
+  m_pipeline_size += total_write_size;
 
-  m_wr_op.reset(m_str->write(buffers(m_wr_buf, net_size + header_length - 1)));
+  if (!m_pipeline)
+  {
+    write(wr_buf);
+  }
+}
+
+
+void Protocol_impl::write(byte *buf)
+{
+  m_wr_op.reset(m_str->write(buffers(buf, m_pipeline_size)));
+  clear_Pipeline();
+}
+
+void Protocol_impl::write()
+{
+  write(m_wr_buf);
 }
 
 
@@ -359,13 +498,39 @@ void Protocol_impl::read_header()
   if (HEADER == m_msg_state)
     return;
 
+  m_msg_state= HEADER;
+
+  if (m_msg_compressed_type)
+  {
+    /*
+      If we are processing compressed messages, and there is more compressed
+      data, uncompress next message header. Otherwise (no more compressed data),
+      get out of compressed mode and proceed to reading next message header
+      from the input stream.
+    */
+
+    if (!m_compressor.uncompression_finished())
+    {
+      if (!m_compressor.uncompress(m_rd_buf, 5))
+        THROW("Error uncompressing the message header");
+      return;
+    }
+    else
+    {
+      m_msg_compressed_type = 0;
+      m_compressor.reset();  // clean up compressor
+    }
+  }
+
+
   if (m_rd_op)
     THROW("can't read header when reading payload is not completed");
 
-  m_rd_op.reset(m_str->read(buffers(m_rd_buf,4)));
-  m_msg_state= HEADER;
+  // Read length and message type
+  m_rd_op.reset(m_str->read(buffers(m_rd_buf, 5)));
 }
 
+
 void Protocol_impl::read_payload()
 {
   if (PAYLOAD == m_msg_state)
@@ -374,56 +539,69 @@ void Protocol_impl::read_payload()
   if (HEADER != m_msg_state)
     THROW("payload can be read only after header");
 
+  m_msg_state = PAYLOAD;
+
+  // Nothing to do if message has no payload.
+
+  if (0 == m_msg_size)
+    return;
+
+  if (!resize_buf(SERVER, m_msg_size))
+    THROW("Not enough memory for input buffer");
+
+  /*
+    If we process compressed data, request compressor to decompress next
+    payload. Otherwise read payload directly from input stream.
+  */
+
+  if (m_msg_compressed_type)
+  {
+    if (!m_compressor.uncompress(m_rd_buf, m_msg_size))
+      THROW("Error uncompressing the message payload");
+    return;
+  }
+
   if (m_rd_op)
     THROW("can't read payload when reading header is not completed");
 
-  if (!resize_buf(SERVER, m_msg_size))
-      THROW("Not enough memory for input buffer");
-
-  if (m_msg_size > 0)
-    m_rd_op.reset(m_str->read(buffers(m_rd_buf, m_msg_size)));
-  m_msg_state= PAYLOAD;
+  m_rd_op.reset(m_str->read(buffers(m_rd_buf, m_msg_size)));
 }
 
 
 bool Protocol_impl::rd_cont()
 {
-  if (!m_rd_op)
-    return true;
+  // First try to finish m_rd_op, if set.
 
-  if (!m_rd_op->cont())
+  if (m_rd_op && !m_rd_op->cont())
     return false;
 
+  // Call rd_process when IO is finished (orthere is no IO to begin with).
+
   m_rd_op.reset();
-
-  if (PAYLOAD == m_msg_state)
-    return true;
-
   rd_process();
 
-  return true;
+  // We are done only if rd_process() did not set up a new IO operation.
+
+  return !m_rd_op;
 }
 
 
 void Protocol_impl::rd_wait()
 {
-  if (m_rd_op)
+  while (!rd_cont())
   {
+    // Note: rd_cont() returns false only if there is pending IO
+    assert(m_rd_op);
     m_rd_op->wait();
-    m_rd_op.reset();
-
-    if (PAYLOAD == m_msg_state)
-      return;
-
-    rd_process();
   }
 }
 
 
 bool Protocol_impl::resize_buf(Protocol_side side, size_t requested_size)
 {
-  byte*  &buf= (side == SERVER ? m_rd_buf : m_wr_buf);
-  size_t &buf_size= (side == SERVER ? m_rd_size : m_wr_size);
+  byte*  &buf = (side == SERVER ? m_rd_buf :m_wr_buf);
+
+  size_t &buf_size = (side == SERVER ? m_rd_size : m_wr_size);
 
   if (requested_size < buf_size)
     return true;
@@ -431,41 +609,119 @@ bool Protocol_impl::resize_buf(Protocol_side side, size_t requested_size)
   // Note that since requested_size >= buf_size, the buffer size is
   // at least doubled here.
 
-  size_t new_size= buf_size + requested_size;
-  byte *ptr= (byte*) realloc(buf, new_size);
+  size_t new_size = buf_size + requested_size;
+  byte *ptr = (byte*)realloc(buf, new_size);
 
   // If allocating buffer with margin failed, try allocating
   // exact required amount.
 
   if (!ptr)
   {
-    new_size= requested_size;
-    ptr= (byte*) realloc(buf, new_size);
+    if (side == CLIENT)
+      new_size = m_pipeline_size + requested_size;
+    else
+      new_size = requested_size;
+    ptr = (byte*)realloc(buf, new_size);
   }
 
   if (!ptr)
     return false;
 
   buf_size = new_size;
-  buf= ptr;
+  buf = ptr;
 
   return true;
 }
 
+#define GET_PAYLOAD_SIZE(S, B) S = *(msg_size_t*)(B); \
+                           NTOHSIZE(S)
+
+
+/*
+  Note: Called from rd_wait() or rd_cont() when the async IO m_rd_op
+  is completed.
+*/
 
 void Protocol_impl::rd_process()
 {
-  m_msg_size= *(msg_size_t*)m_rd_buf;
-  NTOHSIZE(m_msg_size);
-  assert(m_msg_size > 0);
-  m_msg_size--;
+  /*
+    At this point m_rd_op is completed. We have any more
+    work to do here only if we are in HEADER mode and we
+    need to parse the header data that is now available.
+  */
 
-  m_rd_op.reset(m_str->read(buffers(m_rd_buf, 1)));
-  m_rd_op->wait();
-  m_rd_op.reset();
-  m_msg_type= m_rd_buf[0];
+  if (HEADER != m_msg_state)
+    return;
+
+  if (m_msg_compressed_type == 0)
+  {
+    GET_PAYLOAD_SIZE(m_msg_size, m_rd_buf);
+    m_msg_size--;
+    // The read buffer already contains the message type
+    m_msg_type = m_rd_buf[4];
+
+    if (m_msg_type == msg_type::Compression)
+    {
+      m_msg_compressed_type = m_msg_type;
+      // Make sure the reading buffer is large enough
+      if (!resize_buf(SERVER, m_msg_size))
+        THROW("Not enough memory for input buffer");
+
+      m_rd_op.reset(m_str->read(buffers(m_rd_buf, m_msg_size)));
+      m_preamble = true;
+      return;
+    }
+  }
+  else
+  {
+    /*
+      We are processing compressed frame, looking for next message (since
+      m_msg_state is HEADER).
+
+      If compressor was not initialized yet, we are looking at the 5 bytes
+      preamble containing info about compressed data and we can initalize
+      compressor using that data and request uncompressing first payload size.
+
+      Otherwise (compressor already intialized) m_rd_buf contains 4 byte
+      size of the next payload which is stored in m_msg_size.
+    */
+
+    if (m_preamble)
+    {
+      m_preamble = false;
+      m_compressed_msg.Clear();
+      if (!m_compressed_msg.ParseFromArray(m_rd_buf, (int)m_msg_size))
+        throw_error("Invalid Compression message");
+
+      m_compressor.set_compressed_buf((byte*)m_compressed_msg.payload().data(),
+        m_compressed_msg.payload().length(),
+        (size_t)m_compressed_msg.uncompressed_size());
+
+      if (!m_compressor.uncompress(m_rd_buf, 5))
+        throw_error("Error uncompressing the message header");
+
+      GET_PAYLOAD_SIZE(m_msg_size, m_rd_buf);
+      --m_msg_size; // Subtract 1 byte of msg type, which we already know
+      m_msg_type = (msg_type_t)m_rd_buf[4];
+    }
+    else
+    {
+      if (!m_compressor.uncompression_finished())
+      {
+        GET_PAYLOAD_SIZE(m_msg_size, m_rd_buf);
+        --m_msg_size; // Subtract 1 byte of msg type, which we already know
+      }
+    }
+  }
 }
 
+void
+Protocol_impl::set_compression(Compression_type::value compression_type,
+                                    size_t threshold)
+{
+  m_compressor.set_compression_type(compression_type);
+  m_compress_threshold = threshold;
+}
 
 /*
   Processing incoming messages
@@ -588,7 +844,9 @@ bool Op_rcv::do_read_msg(bool async)
         // process the payload
 
         if (m_prc && !m_error)
+        {
           process_payload();
+        }
 
         /*
           call message_end() - the return value can tell us to stop
@@ -657,7 +915,7 @@ void Op_rcv::process_payload()
   assert(m_prc);
   assert(PAYLOAD == m_stage);
 
-  // Send raw message bytes to m_proto if requested (m_read_window > 0).
+  // Send raw message bytes to m_prc if requested (m_read_window > 0).
 
   try {
 
@@ -674,12 +932,11 @@ void Op_rcv::process_payload()
     while (cur_pos < end_pos && m_read_window)
     {
       size_t new_window = m_prc->message_data(bytes(cur_pos,
-                                 cur_pos + m_read_window < end_pos ?
-                                 m_read_window : end_pos - cur_pos));
+                                  cur_pos + m_read_window < end_pos ?
+                                  m_read_window : end_pos - cur_pos));
       cur_pos += m_read_window;
       m_read_window= new_window;
     }
-
     m_prc->message_received(m_msg_size);
 
   }
@@ -801,12 +1058,25 @@ public:
 
     Mysqlx::Ok &ok= static_cast<Mysqlx::Ok&>(msg);
     static_cast<Reply_processor&>(*m_prc).ok(ok.msg());
-  };
+  }
 
 };
 
 
-Protocol::Op& Protocol::snd_Close()
+Protocol::Op& Protocol::snd_SessionReset(bool keep_open)
+{
+  Mysqlx::Session::Reset reset;
+  reset.set_keep_open(keep_open);
+  return get_impl().snd_start(reset, msg_type::cli_SessionReset);
+}
+
+Protocol::Op& Protocol::snd_SessionClose()
+{
+  Mysqlx::Session::Close close;
+  return get_impl().snd_start(close, msg_type::cli_SessionClose);
+}
+
+Protocol::Op& Protocol::snd_ConnectionClose()
 {
   Mysqlx::Connection::Close close;
   return get_impl().snd_start(close, msg_type::cli_Close);
@@ -868,7 +1138,7 @@ public:
 };
 
 
-void Rcv_command::process_msg(msg_type_t type, Message &msg)
+void Rcv_command::process_msg(msg_type_t type, Message&)
 {
   Cmd_processor prc= static_cast<Cmd_processor&>(*m_prc);
 
@@ -885,7 +1155,6 @@ Protocol::Op& Protocol_server::rcv_Command(Cmd_processor &prc)
   return get_impl().rcv_start<Rcv_command>(prc);
 }
 
-
 // ------------------------------------------------------------
 
 size_t Processor_base::message_begin_internal(msg_type_t type, bool &flag)
